@@ -3,13 +3,13 @@
 
 from __future__ import absolute_import, division, print_function, unicode_literals
 
-import os, sys, shutil, argparse, subprocess, unittest
+import os, sys, shutil, argparse, subprocess, unittest, io
 import pexpect, pexpect.replwrap
-from tempfile import TemporaryFile, mkdtemp
+from tempfile import TemporaryFile, NamedTemporaryFile, mkdtemp
 
-TEST_DIR = os.path.abspath(os.path.dirname(__file__))
-BASE_DIR = os.path.dirname(TEST_DIR)
-sys.path.insert(0, BASE_DIR)
+TEST_DIR = os.path.abspath(os.path.dirname(__file__))  # noqa
+BASE_DIR = os.path.dirname(TEST_DIR)  # noqa
+sys.path.insert(0, BASE_DIR)  # noqa
 
 from argparse import ArgumentParser, SUPPRESS
 from argcomplete import (
@@ -17,13 +17,14 @@ from argcomplete import (
     CompletionFinder,
     split_line,
     ExclusiveCompletionFinder,
-    _check_module
+    _check_module,
+    shellcode
 )
 from argcomplete.completers import FilesCompleter, DirectoriesCompleter, SuppressCompleter
 from argcomplete.compat import USING_PYTHON2, str, sys_encoding, ensure_str, ensure_bytes
 
 IFS = "\013"
-COMP_WORDBREAKS = " \t\n\"'><=;&(:"
+COMP_WORDBREAKS = " \t\n\"'><=;|&(:"
 
 BASH_VERSION = subprocess.check_output(['bash', '-c', 'echo $BASH_VERSION']).decode()
 BASH_MAJOR_VERSION = int(BASH_VERSION.split('.')[0])
@@ -69,8 +70,7 @@ class TestArgcomplete(unittest.TestCase):
         command = ensure_str(command)
 
         if point is None:
-            # Adjust point for wide chars
-            point = str(len(command.encode(sys_encoding)))
+            point = str(len(command))
         with TemporaryFile() as t:
             os.environ["COMP_LINE"] = ensure_bytes(command) if USING_PYTHON2 else command
             os.environ["COMP_POINT"] = point
@@ -107,7 +107,10 @@ class TestArgcomplete(unittest.TestCase):
             ("prog --shi", ["--ship "]),
             ("prog --ship ", ["submarine", "speedboat"]),
             ("prog --ship s", ["submarine", "speedboat"]),
-            ("prog --ship su", ["submarine "])
+            ("prog --ship su", ["submarine "]),
+            ("prog --ship=", ["submarine", "speedboat"]),
+            ("prog --ship=s", ["submarine", "speedboat"]),
+            ("prog --ship=su", ["submarine "]),
         )
 
         for cmd, output in expected_outputs:
@@ -681,6 +684,7 @@ class TestArgcomplete(unittest.TestCase):
         expected_outputs = (
             ("prog ", ["--get", "--set", "name1", "name2"]),
             ("prog --", ["--get", "--set"]),
+            ("prog -- ", ["name1", "name2"]),
             ("prog --get ", ["--get", "name1", "name2"]),
             ("prog --get name1 ", ["--get "])
         )
@@ -738,6 +742,20 @@ class TestArgcomplete(unittest.TestCase):
         self.assertEqual(set(self.run_completer(make_parser(), 'prog -3 "')), {"\"'"})
         self.assertEqual(set(self.run_completer(make_parser(), "prog -3 '")), {"\"'"})
 
+    def test_shellcode_utility(self):
+        with NamedTemporaryFile() as fh:
+            sc = shellcode(["prog"], use_defaults=True, shell="bash", complete_arguments=None)
+            fh.write(sc.encode())
+            fh.flush()
+            subprocess.check_call(['bash', '-n', fh.name])
+        with NamedTemporaryFile() as fh:
+            sc = shellcode(["prog", "prog2"], use_defaults=False, shell="bash", complete_arguments=["-o", "nospace"])
+            fh.write(sc.encode())
+            fh.flush()
+            subprocess.check_call(['bash', '-n', fh.name])
+        sc = shellcode(["prog"], use_defaults=False, shell="tcsh", complete_arguments=["-o", "nospace"])
+        sc = shellcode(["prog"], use_defaults=False, shell="woosh", complete_arguments=["-o", "nospace"])
+        sc = shellcode(["prog"], shell="fish")
 
 class TestArgcompleteREPL(unittest.TestCase):
     def setUp(self):
@@ -979,6 +997,7 @@ class _TestSh(object):
     """
     sh = None
     expected_failures = []
+    skipped = []
 
     @classmethod
     def setUpClass(cls, *args, **kwargs):
@@ -989,6 +1008,13 @@ class _TestSh(object):
             def wrapped(self, test=test):
                 test(self)
             setattr(cls, name, wrapped)
+        for name in cls.skipped:
+            test = getattr(cls, name)
+
+            @unittest.skip('skip')
+            def wrapped(self, test=test):
+                pass
+            setattr(cls, name, wrapped)
         super(_TestSh, cls).setUpClass(*args, **kwargs)
 
     def setUp(self):
@@ -1076,6 +1102,14 @@ class _TestSh(object):
     def test_completion_environment(self):
         self.assertEqual(self.sh.run_command('prog env o\t'), 'ok\r\n')
 
+    def test_comp_point(self):
+        # Use environment variable to change how prog behaves
+        self.assertEqual(self.sh.run_command('export POINT=1'), '')
+        self.assertEqual(self.sh.run_command('prog point hi\t'), '13\r\n')
+        self.assertEqual(self.sh.run_command('prog point hi \t'), '14\r\n')
+        self.assertEqual(self.sh.run_command('prog point 你好嘚瑟\t'), '15\r\n')
+        self.assertEqual(self.sh.run_command('prog point 你好嘚瑟 \t'), '16\r\n')
+
 
 class TestBash(_TestSh, unittest.TestCase):
     expected_failures = [
@@ -1148,6 +1182,42 @@ class TestBashGlobal(TestBash):
             self.sh.run_command('cd ' + os.getcwd())
             self.assertEqual(self.sh.run_command('python -m package.prog basic f\t'), 'foo\r\n')
 
+    def _test_console_script(self, package=False, wheel=False):
+        with TempDir(prefix='test_dir_py', dir='.'):
+            self.sh.run_command('cd ' + os.getcwd())
+            self.sh.run_command('export PATH=$PATH:./bin')
+            self.sh.run_command('export PYTHONPATH=.:$PYTHONPATH')
+            test_package = os.path.join(TEST_DIR, 'test_package')
+            command = 'pip install {} --target .'.format(test_package)
+            if not wheel:
+                command += ' --no-binary :all:'
+                if sys.platform == 'darwin':
+                    # Work around https://stackoverflow.com/questions/24257803
+                    command += ' --install-option="--prefix="'
+            install_output = self.sh.run_command(command)
+            self.assertEqual(self.sh.run_command('echo $?'), '0\r\n', install_output)
+            command = 'test-module'
+            if package:
+                command = 'test-package'
+            command += ' a\t'
+            self.assertEqual(self.sh.run_command(command), 'arg\r\n')
+
+    def test_console_script_module(self):
+        """Test completing a console_script for a module."""
+        self._test_console_script()
+
+    def test_console_script_package(self):
+        """Test completing a console_script for a package."""
+        self._test_console_script(package=True)
+
+    def test_console_script_module_wheel(self):
+        """Test completing a console_script for a module from a wheel."""
+        self._test_console_script(wheel=True)
+
+    def test_console_script_package_wheel(self):
+        """Test completing a console_script for a package from a wheel."""
+        self._test_console_script(package=True, wheel=True)
+
 
 class TestTcsh(_TestSh, unittest.TestCase):
     expected_failures = [
@@ -1156,6 +1226,8 @@ class TestTcsh(_TestSh, unittest.TestCase):
         'test_continuation',
         'test_parse_special_characters',
         'test_parse_special_characters_dollar',
+        # Test case doesn't work under tcsh, could be fixed.
+        'test_comp_point',
     ]
 
     def setUp(self):
@@ -1175,6 +1247,36 @@ class TestTcsh(_TestSh, unittest.TestCase):
             self.sh.run_command('')
 
 
+class TestFish(_TestSh, unittest.TestCase):
+    expected_failures = [
+        'test_parse_special_characters',
+        'test_comp_point',
+        'test_special_characters_double_quoted'
+    ]
+
+    skipped = [
+        'test_single_quotes_in_single_quotes',
+        'test_parse_special_characters_dollar'
+    ]
+
+    def setUp(self):
+
+        sh = Shell('fish')
+        path = ' '.join([os.path.join(BASE_DIR, 'scripts'), TEST_DIR, '$PATH'])
+        sh.run_command('set -x PATH {0}'.format(path))
+        sh.run_command('set -x PYTHONPATH {0}'.format(BASE_DIR))
+        output = sh.run_command('register-python-argcomplete --shell fish prog | .')
+        self.assertEqual(output, '')
+        self.sh = sh
+
+    def tearDown(self):
+        # The shell wrapper is fragile; exactly which exception is raised
+        # differs depending on environment.
+        with self.assertRaises((pexpect.EOF, OSError)):
+            self.sh.run_command('exit')
+            self.sh.run_command('')
+
+
 class Shell(object):
     def __init__(self, shell):
         self.child = pexpect.spawn(shell, encoding='utf-8')
